package it.jrc.entitymatcher;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.zip.GZIPInputStream;

import java.io.Reader;
import java.io.Writer;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;

import it.jrc.matcher.MatchWords;
import it.jrc.matcher.MatchResult;

public class JRCNames
{

	private MatchWords matchWords = MatchWords.getInstance("entityMatcher");
	private ArrayList<String> code2lan;

	/**
	 * Method JRCNames
	 *
	 *
	 */
	public JRCNames ()
	{
		// nothing much to do in this constructor
	}

	/**
	 *	Parse the entities file.
	 *
	 *	@param br	is a buffered reader for the entities file that allows us to read the file line by line
	 *	@param entities is a Hashtable where the information structure associated with each entity is stored.
	 *	@param names contains the 'preferred display name' for each entity ID
	 * 	@param code2lan maps stores the language code. The INDEX of the language code is stored in the entity information structure.
	 *
	 *	@throws an exception... IO or otherwise... purists will have a heart attack, but it simply throws an exception
	 */
	public void initKnownEntities(BufferedReader br,
														Hashtable<String, StringBuilder> entities,
														HashMap<Integer, String> names,
														ArrayList<String>	code2lan) throws Exception
	{
		int nDuplicates = 0;

		String line = br.readLine();
		while (line != null && !line.equals(""))
		{
			Entity e = new Entity();
			e.init(line);
			if (e.alias == null || e.alias.equals(""))
			{
				System.out.println("line "+line+" does not contain alias");
			}
			else
			{
				// ok, add this alias
				if (e.alias.length() > 0)
				{
					String language = e.lan;
					e.languageCode = code2lan.indexOf(language);
					if (e.languageCode == -1)
					{
						e.languageCode = code2lan.size();
						code2lan.add(language);
					}

					char[] code = e.charCode();
					StringBuilder sid = entities.get(e.alias);
					boolean match = false;
					if (sid == null)
					{
						sid = new StringBuilder();
						sid.append('\1');
					}
					else
					{
						// we should check that this new alias+attributes is unique ...
						// TODO: This check can be renmoved when we have a properly constructed input file...
						int nCount = sid.charAt(0);
						int i = 0;
						while (i<nCount && !match)
						{
							int c=0;
							while (c < 4 && !match) {match = code[c] == sid.charAt(i*4+1+c); c++;}
							i++;
						}
						if (!match) sid.setCharAt(0, (char)(nCount+1));
					}
					if (!match)
					{
						sid.append(code);
						entities.put(e.alias, sid);
						if (names.get(e.id) == null) names.put(e.id, e.alias.replaceAll("\\+", " "));
					}
					else
					{
						//System.out.println("Warning: found duplicate entry for "+e.alias+": id= "+e.id+" language code = "+e.languageCode+" type="+e.type);
						nDuplicates++;
					}
				}
				else
				{
					System.out.println("id found, but alias is blank "+line);
				}
			}
			// do the next line
			line = br.readLine();

		}
		if (nDuplicates > 0) System.out.println("found "+nDuplicates+" duplicate alias/languagecode/type combinations!!!!");
	}

	/**
	 *	Copy reader to writer
	 */
	private void copy(Reader in, Writer out) throws IOException
  {
    // Read and write by chunks until eof
    int chunkSize = 4*4096;
    char[] ca = new char[chunkSize];
    int charsRead = in.read(ca, 0, chunkSize);
    while(charsRead >= 0)
    {
      out.write(ca, 0 , charsRead);
      charsRead = in.read(ca, 0, chunkSize);
    }
    in.close();
    out.flush();
    out.close();
  }


	/**
	 * Method main
	 *
	 *
	 * 	@param entity file, the gzip file containing the JRC names definition
	 *	@param text file, utf-8 encoded file containing the text to be parsed
	 *	@param language code. The language code that corresponds to the text file language
	 *
	 */
	public static void main (String[] args) throws Exception
	{

		int nArgs = 0;
		// some basic checking on the arguments. We need 2, or 3 of them
		if (args.length == 2)
		{
			nArgs = 2;
		}
		else if (args.length == 3)
		{
			nArgs = 3;
		}
		else
		{
			System.out.println(" To detect entities in a text");
			System.out.println("JRCNames file (the entities.gzip) file (the text to parse) language (2 character iso language code for the text) ");
			System.out.println(" ------ OR --------");
			System.out.println(" To find all known name variants for a name");
			System.out.println("JRCNames file (the entities.gzip) \"name\" ");
		}

		if (nArgs > 0)
		{

			JRCNames jrcNames = new JRCNames();

			MatchWords matchWords = MatchWords.getInstance("entityMatcher");
			// construct datastructures for matcher (and names for stats/display)
	    Hashtable<String, StringBuilder> entityTable = new Hashtable<String, StringBuilder>();
			HashMap<Integer, String> names = new HashMap<Integer, String>();
			ArrayList<String> code2lan = new ArrayList<String>();
	   	// PRESET 'u' to 0!!!!!
	    code2lan.add("u");



			// the entities file
			String entitiesFilename = args[0];

			GZIPInputStream zis = new GZIPInputStream(new FileInputStream(entitiesFilename));
			BufferedReader br = new BufferedReader(new InputStreamReader(zis, "utf-8"));
			jrcNames.initKnownEntities(br, entityTable, names, code2lan);
	  	System.out.println("found "+entityTable.size()+" aliases");
	  	matchWords.init(entityTable);

			if (nArgs == 2)
			{
				// we can use the entity table to look up the stuff...
				String theName = args[1];
				// hard code remove first and last character (should be ")
				String text = theName;
				System.out.println("looking for "+text);

				EntityMatchResult result = new EntityMatchResult(code2lan.indexOf("u"));
				matchWords.match(text, result);
				for (Integer id: result.getEntities())
				{
					System.out.println("found entity ID "+id);
					char[] charId = Entity.charID(id);
					System.out.println("looking through "+entityTable.size()+" names...");
					// go and look for these name variants
					for (String name:entityTable.keySet())
					{
						StringBuilder sb = entityTable.get(name);

						if ((sb.charAt(1) == charId[0]) && (sb.charAt(2) == charId[1])) System.out.println(name.replaceAll("\\+", " "));
					}
				}

			}
			else
			{

		  	// the data has now been copied into the Matcher's data structure and can be deleted
		  	entityTable.clear();


				// the text to parse
				String textFile = args[1];
				InputStreamReader r = new InputStreamReader(new FileInputStream(textFile), "utf-8");
				StringWriter w = new StringWriter();
				jrcNames.copy(r,w);
				String text = w.toString();

				System.out.println();
				System.out.println(text);

				// the language of the text
				String language = args[2];
				EntityMatchResult result = new EntityMatchResult(code2lan.indexOf(language));
				matchWords.match(text, result);

				for (Integer id: result.getEntities())
				{
					ArrayList<String> aliases = result.getAliases(id);
					ArrayList<Integer> positions = result.getCharIndexes(id);
					String name = names.get(id);
					char type = result.getType(id);
					for (int j=0; j<aliases.size(); j++)
					{
						String alias = aliases.get(j);
						if (alias != null)
						{
							StringBuilder pos = new StringBuilder();
							pos.append(positions.get(j));
							int count = 1;
							for (int k=j+1; k<aliases.size(); k++)
							{
								String next = aliases.get(k);
								if (next!=null && next.equals(alias))
								{
									count++;
									pos.append(",").append(positions.get(k));
									aliases.set(k, null);
								}
							}
							System.out.println("found entity id = "+id+" type "+Character.toString(type)+" as "+alias+" ("+pos.toString()+")");
						}
					}

				}
			}
		}
	}
}
